/*demo_fluttercode.cpp*/
/*
 * Copyright 2012 IPOL Image Processing On Line http://www.ipol.im/
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @mainpage Flutter shutter code optimizer
 *
 * README.txt:
 * @verbinclude README.txt
 */

/**
 * @file demo_fluttercode.cpp
 * @brief Flutter shutter code optimizer main file
 *
 *
 * @author Yohann Tendero <yohann.tendero@cmla.ens-cachan.fr>
 */

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <math.h>
#include <fstream>





#ifndef M_PI
/**
 * M_PI is a POSIX definition for Pi
 */
#define M_PI 3.14159265358979323846
#endif

#include "flutter_optimizer_gaussian.h"
#include "flutter_optimizer_uniform.h"
#include "flutter_optimizer_handcrafted.h"
#include "code_comparison.h"
#include "best_snapshot_average.h"
using namespace std;

/**
 * @brief main function call
 *
INPUTs:
*Motion model : 0 centered Gaussian (0,s) => argv1
1 centered Uniform
* the s parameter.
*N : length of the code
*Deltat : time steps
 Outputs
 txt file containing the a_k.
 EPS file containing the stairs a_k
 EPS file containing the FT of the flutter function
 EPS file containing the comparativ relative_efficiency of the coda  and the
 corresoinding snapshot for v_max.


 *
 */
int main(int argc, char **argv)
{
//Check arguments : IN OUT;
    if (argc < 9)
    {
        std::cerr << " ***************************************** " << std::endl
                  << " **********  Flutter shutter code calculator  ***** " << std::endl
                  << " *************************************************** " << std::endl
                  << " Usage:  input : motion _model and parameter. "<< std::endl
                  << " Integration time factor  multiplication, code_length." << std::endl
                  << " *************************************************** " << std::endl
                  << " *************  Yohann Tendero, 2012  ************** " << std::endl
                  << " *************************************************** " << std::endl;
        return 1;
    }


    /**
    * @brief : Overall description:
    * Step 1 : read parameter;
    * Step 2 : Given the motion model : compute best snapshot on average
    * (the best exposure time to use in a standard camera to ensure the best SNR
    * taking the deconvolution into account);
    * Step 3 : Given the motion model : compute the best flutter shutter
    * camera design (the best aperture strategy);
    * Step 4-5 : store results.
    *
    **/




    ////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////
    //// Step1. Reading arguments, init ////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////

    int flag_motion_model=atoi(argv[1]);
    /// 0 For Gaussian, 1 for uniform, 2 for handcrafted
    double s=atof(argv[2]); // motion model parameter
    /// motion model is N(0,s) if Gaussian and [-s,s] if uniform.
    int code_length=atoi(argv[3]); // length of the code L
    double  exposure_factor=atof(argv[4]);// tune the support of the
    // flutter shutter function as a factor of the best snapshot
    // exposure time.
    double* code = new double[code_length];  //to store the flutter shutter code
    const int num_plot_points=4000;  // number of points for plots
    double best_snapshot_deltat; // to store the best exposure time (std camera)
    double v_max=s;// maximum velocity.
    if (flag_motion_model==0) v_max=4*s;
    double proba_handcrafted_0=0; // for handcrafted motion model only
    // storing files names for output.
    char* code_file_name=argv[5];
    char* TF_code_file_name=argv[6];
    char* SNR_comparison_file_name=argv[7];
    char* mean_sdt_file_name=argv[8];

    if(flag_motion_model==2) // handcrafted motion model
    {
        proba_handcrafted_0=atof(argv[5]);
        code_file_name=argv[6];
        TF_code_file_name=argv[7];
        SNR_comparison_file_name=argv[8];
        mean_sdt_file_name=argv[9];
    }

    ////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////
    //// Step2 Best snapshot average computation  //////////////////////////////
    ////////////////////////////////////////////////////////////////////////////

    best_snapshot_deltat=best_snapshot_average(flag_motion_model, s,
                         proba_handcrafted_0);
    double deltat=exposure_factor*best_snapshot_deltat/code_length;

    /*std::cerr <<  "Best snapshot exposure time" << std::endl;
    std::cerr <<  best_snapshot_deltat << std::endl;
    std::cerr <<  "Maximum blur for best snapshot" << std::endl;
    std::cerr <<  best_snapshot_deltat*v_max << std::endl;
    std::cerr <<  "|v_{max}|  Delta t" << std::endl;
    std::cerr <<  deltat*v_max << std::endl;*/
    if (deltat*v_max>1)
    {
        std::cerr << " Deltat too big! convergence not guaranteed" << std::endl;
        std::cerr << " Increase the code length L."  << std::endl;
    }

    ////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////
    //// Step3 Code computation  ///////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////


    if (flag_motion_model==0) //Gaussian motion model
    {
        gaussian_optimisator(code,s,code_length,deltat);
    }
    if (flag_motion_model==1)
    {
        uniform_optimisator(code,s,code_length,deltat);
    }
    if (flag_motion_model==2)
    {
        handcrafted_optimisator(code,s,proba_handcrafted_0,code_length,deltat);
    }




    ////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////
    //// Step4 : Writing the code file /////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////


    ///WRITE the code to file : code_file_name
    ofstream file_code(code_file_name, ios::out | ios::trunc);
    // opening the file (erase if previous exists)
    if (file_code)
    {
        for (int k=0; k<code_length; k++)
        {
            file_code << code[k] <<";" << endl;
        }

        file_code.close();
    }
    else
        cerr << "Unable to open file !" << endl;

    ///WRITE the Fourier transfor of the  code to file : TF_code_file_name :
    ///This one contains (columns by columns)
    ///     \xi for \xi \in [-\pi v_max \deltat, \pi \vmax \deltat]
    /// 	The ideal Fourier Transform
    ///	    The Fourier transform of the code




    ///  Flutter shutter codes are defined up to a constant renormalization,
    /// the following renormalize so \hat \alpha(0)=\sqrt[4] \W(0)=1.
    double w_function_0=0;
    if (flag_motion_model==0) w_function_0=gaussian_w(s,0);
    if (flag_motion_model==1) w_function_0=uniform_w(s,0);
    if (flag_motion_model==2) w_function_0=handcrafted_w(0,proba_handcrafted_0,
                                                         deltat,0);
    double hat_alpha_0=abs_hat_alpha(code, code_length, 0,deltat);


    ofstream file_TFcode(TF_code_file_name, ios::out | ios::trunc);
    if (file_TFcode)
    {
        if (flag_motion_model==0)
            for (double xi=-2*v_max*M_PI; xi<=2*v_max*M_PI;
                    xi=xi+4*v_max*M_PI/num_plot_points)
            {
                file_TFcode << xi
                << " "
                << pow(gaussian_w( s,  xi)/w_function_0,0.25)
                <<" "
                << abs_hat_alpha(code, code_length, xi,deltat)
                /hat_alpha_0 << endl;
            }
        if (flag_motion_model==1)
            for (double xi=-2*v_max*M_PI; xi<=2*v_max*M_PI;
                    xi=xi+4*v_max*M_PI/num_plot_points)
            {
                file_TFcode << xi
                << " "
                << pow(uniform_w( s,  xi)/w_function_0,0.25)
                <<" "
                << abs_hat_alpha(code, code_length, xi,deltat)
                /hat_alpha_0 << endl;
            }

        if (flag_motion_model==2)
            for (double xi=-2*v_max*M_PI; xi<=2*v_max*M_PI;
                    xi=xi+4*v_max*M_PI/num_plot_points)
            {
                file_TFcode << xi
                << " "
                << pow(handcrafted_w(s, proba_handcrafted_0,deltat,
                                     xi*deltat)/w_function_0,0.25)
                <<" "
                << abs_hat_alpha(code, code_length, xi,deltat)
                /hat_alpha_0 << endl;
            }

        file_TFcode.close();
    }
    else
        cerr << "Unable to open file !" << endl;

    ////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////
    //// Step5 : Code comparison with snapshot  ////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////
    // Compute the relative efficiency, the relative efficiency on average,
    // the risk and store them.

    int i=0;
    double* relative_efficiency = new double[num_plot_points];
    double* relative_efficiency_on_average = new double[1];
    double* risk = new double[1];;
    if (flag_motion_model==0) v_max=4*s;
    averaged_relative_efficiency( code,  code_length,  best_snapshot_deltat,
                                  flag_motion_model, s,  deltat,num_plot_points,
                                  relative_efficiency,
                                  relative_efficiency_on_average,risk,
                                  proba_handcrafted_0);
    ofstream file_comparison_file_name(SNR_comparison_file_name,
                                       ios::out | ios::trunc);
// opening the file (erase if previous exists)
    if (file_comparison_file_name)
    {
        for (double v=-v_max; v<=v_max; v=v+2*v_max/num_plot_points)
        {
            file_comparison_file_name << v
            << " "
            <<  relative_efficiency[i]
            << " "
            <<  relative_efficiency_on_average[0]
            << " "
            <<  endl;
            i++;

        }

        file_comparison_file_name.close();
    }
    else
        cerr << "Unable to open file !" << endl;

    std::cerr <<  "Maximum blur for best snapshot" << std::endl;
    std::cerr <<  best_snapshot_deltat*v_max << std::endl;
    printf("Signal to noise ratio increase (on average): %f\n",
           relative_efficiency_on_average[0]);
    printf(" Risk: %f\n", risk[0]);



    ofstream file_mean_sdt_file_name(mean_sdt_file_name, ios::out | ios::trunc);
// opening the file (erase if previous exists)
    if (file_mean_sdt_file_name)
    {

        file_mean_sdt_file_name
        << "Signal to noise ratio increase summary" << std::endl
        << "*Relative efficency on average: " <<
        relative_efficiency_on_average[0] << std::endl
        <<"*Risk: " << risk[0] <<   endl;
        file_mean_sdt_file_name.close();
    }
    else
        cerr << "Unable to open file !" << endl;

    delete [] code;
    delete [] relative_efficiency;
    delete [] relative_efficiency_on_average;
    delete [] risk;
    return 0;
}




